#!/usr/bin/env python

import argparse
import pandas as pd
import json
import sys
import os

from tabulate import tabulate
from dcm_check import (
    load_ref_json,
    load_ref_dicom,
    load_ref_pydantic,
    get_compliance_summary,
    load_dicom
)

def infer_type_from_extension(ref_path):
    """Infer the reference type based on the file extension."""
    _, ext = os.path.splitext(ref_path.lower())
    if ext == ".json":
        return "json"
    elif ext in [".dcm", ".IMA"]:
        return "dicom"
    elif ext == ".py":
        return "pydantic"
    else:
        print("Error: Could not determine the reference type. Please specify '--type'.", file=sys.stderr)
        sys.exit(1)

def main():
    parser = argparse.ArgumentParser(description="Check DICOM compliance against a reference model.")
    parser.add_argument("--ref", required=True, help="Reference JSON file, DICOM file, or Python module to use for compliance.")
    parser.add_argument("--type", choices=["json", "dicom", "pydantic"], help="Reference type: 'json', 'dicom', or 'pydantic'.")
    parser.add_argument("--acquisition", required=False, help="Acquisition name when using a JSON or Pydantic reference.")
    parser.add_argument("--series", required=False, help="Specific series name within the acquisition for JSON references.")
    parser.add_argument("--in", dest="in_file", required=True, help="Path to the DICOM file to check.")
    parser.add_argument("--fields", nargs="*", help="Optional: List of DICOM fields to include in validation for DICOM reference.")
    parser.add_argument("--out", required=False, help="Path to save the compliance report in JSON format.")
    args = parser.parse_args()

    # Determine the reference type if not provided
    ref_type = args.type or infer_type_from_extension(args.ref)

    if ref_type == "json":
        # Include series if specified
        if args.series:
            reference_model = load_ref_json(args.ref, args.acquisition, series_name=args.series)
        else:
            reference_model = load_ref_json(args.ref, args.acquisition)
    elif ref_type == "dicom":
        ref_dicom_values = load_dicom(args.ref)
        reference_model = load_ref_dicom(ref_dicom_values, args.fields)
    elif ref_type == "pydantic":
        # check if acquisition is provided
        if not args.acquisition:
            print("Error: Acquisition type is required (--acquisition) when using a Pydantic reference.", file=sys.stderr)
            sys.exit(1)
        reference_model = load_ref_pydantic(args.ref, args.acquisition)
    else:
        print(f"Error: Unsupported reference type '{ref_type}'", file=sys.stderr)
        sys.exit(1)

    in_dicom_values = load_dicom(args.in_file)
    results = get_compliance_summary(reference_model, in_dicom_values, args.acquisition, args.series)

    df = pd.DataFrame(results)

    # remove "Acquisition" and/or "Series" columns if they are empty
    if "Acquisition" in df.columns and df["Acquisition"].isnull().all():
        df.drop(columns=["Acquisition"], inplace=True)
    if "Series" in df.columns and df["Series"].isnull().all():
        df.drop(columns=["Series"], inplace=True)

    if len(df) == 0:
        print("DICOM file is compliant with the reference model.")
    else:
        print(tabulate(df, headers="keys", tablefmt="simple"))

    if args.out:
        with open(args.out, 'w') as outfile:
            json.dump(results, outfile, indent=4)
        print(f"Compliance report saved to {args.out}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
import json
import argparse
import pandas as pd
from tabulate import tabulate
from dcm_check import load_ref_json, load_dicom, get_compliance_summary, read_session, interactive_mapping

def get_compliance_summaries_json(json_ref: str, in_session: str, output_json: str = "compliance_report.json", interactive=True) -> pd.DataFrame:
    """
    Generate a compliance summary for each matched acquisition in an input DICOM session.

    Args:
        json_ref (str): Path to the JSON reference file.
        in_session (str): Directory path for the DICOM session.
        output_json (str): Path to save the JSON compliance summary report.

    Returns:
        pd.DataFrame: Compliance summary DataFrame.
    """
    # Step 1: Identify matched acquisitions and series in the session
    session_df, acquisitions_info = read_session(json_ref, in_session, return_acquisitions_info=True)
    grouped_compliance = {}

    # Step 2: Interactive mapping adjustment (if enabled by the user)
    if sys.stdin.isatty() and interactive:
        print("Entering interactive mapping mode. Use arrow keys to navigate, Enter to select and move, and Esc to finish.")
        session_df = interactive_mapping(session_df, acquisitions_info)

    # Step 3: Iterate over each matched acquisition-series pair
    for _, row in session_df.dropna(subset=["Acquisition"]).iterrows():
        acquisition = row["Acquisition"]
        series = row["Series"]
        first_dicom_path = row["First_DICOM"]
        
        try:
            # Load the reference model for the matched acquisition and series
            reference_model = load_ref_json(json_ref, acquisition, series)
            # Load DICOM values for the first DICOM in the series
            dicom_values = load_dicom(first_dicom_path)
            # Run compliance check and gather results
            compliance_summary = get_compliance_summary(reference_model, dicom_values, acquisition, series)

            # Organize results in nested format without "Model_Name"
            if acquisition not in grouped_compliance:
                grouped_compliance[acquisition] = {"Acquisition": acquisition, "Series": []}
            
            if series:
                series_entry = next((g for g in grouped_compliance[acquisition]["Series"] if g["Name"] == series), None)
                if not series_entry:
                    series_entry = {"Name": series, "Parameters": []}
                    grouped_compliance[acquisition]["Series"].append(series_entry)
                for entry in compliance_summary:
                    entry.pop("Acquisition", None)
                    entry.pop("Series", None)
                series_entry["Parameters"].extend(compliance_summary)
            else:
                # If no series, add parameters directly under acquisition
                for entry in compliance_summary:
                    entry.pop("Acquisition", None)
                    entry.pop("Series", None)
                grouped_compliance[acquisition]["Parameters"] = compliance_summary

        except Exception as e:
            print(f"Error processing acquisition '{acquisition}' and series '{series}': {e}")

    # Convert the grouped data to a list for JSON serialization
    grouped_compliance_list = list(grouped_compliance.values())

    # Save grouped compliance summary to JSON
    with open(output_json, "w") as json_file:
        json.dump(grouped_compliance_list, json_file, indent=4)

    # Check if there are any compliance issues to report (ie. if the 'Parameters' list is empty for all acquisitions)
    compliance_issues = any(acq.get("Parameters") or any(series.get("Parameters") for series in acq.get("Series", [])) for acq in grouped_compliance_list)
    print(compliance_issues)
    if not compliance_issues:
        return pd.DataFrame(columns=["Acquisition", "Series", "Parameter", "Value", "Expected"])
    
    # Step 6: Normalize into DataFrame
    df_with_series = pd.json_normalize(
        grouped_compliance_list,
        record_path=["Series", "Parameters"],
        meta=["Acquisition", ["Series", "Name"]],
        errors="ignore"
    )
    df_with_series.rename(columns={"Series.Name": "Series"}, inplace=True)
    df_with_series = df_with_series[["Acquisition", "Series", "Parameter", "Value", "Expected"]]

    # Normalize acquisitions without series directly
    df_without_series = pd.json_normalize(
        [acq for acq in grouped_compliance_list if "Parameters" in acq],
        record_path="Parameters",
        meta=["Acquisition"],
        errors="ignore"
    )
    df_without_series.insert(1, "Series", None)  # Add Series column with None values
    df_without_series = df_without_series[["Acquisition", "Series", "Parameter", "Value", "Expected"]]

    # Combine both DataFrames
    compliance_df = pd.concat([df_with_series, df_without_series], ignore_index=True)

    return compliance_df

def main():
    parser = argparse.ArgumentParser(description="Generate compliance summaries for a DICOM session based on JSON reference.")
    parser.add_argument("--json_ref", required=True, help="Path to the JSON reference file.")
    parser.add_argument("--in_session", required=True, help="Directory path for the DICOM session.")
    parser.add_argument("--output_json", default="compliance_report.json", help="Path to save the JSON compliance summary report.")
    parser.add_argument("--auto_yes", action="store_true", help="Automatically map acquisitions to series.")
    args = parser.parse_args()

    # Generate compliance summaries with interactive mapping
    compliance_df = get_compliance_summaries_json(args.json_ref, args.in_session, args.output_json, not args.auto_yes)

    # if compliance_df is empty, print message and exit
    if compliance_df.empty:
        print("Session is fully compliant with the reference model.")
        return
    
    # Print formatted output with tabulate
    print(tabulate(compliance_df, headers="keys", tablefmt="simple"))

if __name__ == "__main__":
    main()



from pydantic import BaseModel, Field, field_validator, model_validator, confloat, model_validator
from pydantic_core import PydanticCustomError
from typing import List
import numpy as np

class T1_MPR_Config(BaseModel):
    SeriesDescription: str
    MagneticFieldStrength: confloat(ge=2.9, le=3.1)
    RepetitionTime: confloat(ge=2300, le=2500)
    EchoTime: float
    PixelSpacing: List[float]
    SliceThickness: confloat(ge=0.75, le=0.85)

    @field_validator("SeriesDescription")
    def validate_series_description(cls, v):
        if "t1" not in v.lower():
            raise ValueError("SeriesDescription must contain 'T1'")
        return v

    @field_validator("PixelSpacing", mode="before")
    def validate_pixel_spacing(cls, v):
        if len(v) != 2:
            raise ValueError("PixelSpacing must be a list of exactly 2 values")
        if not all(0.75 <= val <= 0.85 for val in v):
            raise ValueError("Each value in PixelSpacing must be between 0.75 and 0.85")
        return v
    
    @model_validator(mode="after")
    def check_repetition_vs_echo_time(self):
        rt, et = self.RepetitionTime, self.EchoTime
        if rt is not None and et is not None and rt < 2 * et:
            raise PydanticCustomError(
                "value_error",
                "RepetitionTime must be at least 2x EchoTime",
                {"RepetitionTime": rt, "EchoTime": et}
            )
        return self

class T2w_SPC_Config(BaseModel):
    SeriesDescription: str
    MagneticFieldStrength: confloat(ge=2.9, le=3.1)
    RepetitionTime: confloat(ge=3.1, le=3.3)
    EchoTime: confloat(ge=0.054, le=0.058)
    PixelSpacing: List[confloat(ge=0.75, le=0.85)]
    SliceThickness: confloat(ge=0.75, le=0.85)

    @field_validator("SeriesDescription")
    def validate_series_description(cls, v):
        if v != "T2w_SPC":
            raise ValueError("SeriesDescription must be 'T2w_SPC'")
        return v

    @field_validator("PixelSpacing")
    def validate_pixel_spacing(cls, v):
        if len(v) != 2:
            raise ValueError("PixelSpacing must have exactly 2 values")
        return v

class DiffusionConfig(BaseModel):
    SeriesDescription: str
    bValue: int = Field(..., ge=900, le=1100)
    NumberOfDirections: int = Field(..., ge=6)
    PixelSpacing: List[confloat(ge=1.15, le=1.35)] = [1.25, 1.25]
    SliceThickness: confloat(ge=1.15, le=1.35) = 1.25

    @field_validator("SeriesDescription")
    def validate_series_description(cls, v):
        if not v.startswith("Diff_"):
            raise ValueError("SeriesDescription must start with 'Diff_'")
        return v

    @field_validator("PixelSpacing")
    def validate_pixel_spacing(cls, v):
        if len(v) != 2:
            raise ValueError("PixelSpacing must have exactly 2 values")
        return v

# Dictionary to map acquisitions to their respective config models (without instantiation)
ACQUISITION_MODELS = {
    "T1_MPR": T1_MPR_Config,
    "T2w_SPC": T2w_SPC_Config,
    "Diff_1k": DiffusionConfig,
}




import pytest
from dcm_check import load_ref_pydantic, is_compliant, get_compliance_summary

@pytest.fixture
def t1_mpr_dicom_values():
    """Fixture for T1_MPR DICOM test values."""
    return {
        "SeriesDescription": "t1_mpr_sag",
        "MagneticFieldStrength": 3.0,
        "RepetitionTime": 2400,
        "EchoTime": 12,
        "PixelSpacing": [0.8, 0.8],
        "SliceThickness": 0.8
    }

@pytest.fixture
def invalid_t1_mpr_dicom_values():
    """Fixture for invalid T1_MPR DICOM values."""
    return {
        "SeriesDescription": "t2",  # Does not contain "T1"
        "MagneticFieldStrength": 2.0,  # Out of range
        "RepetitionTime": 1000,  # Too low, fails repetition vs echo rule
        "EchoTime": 700,
        "PixelSpacing": [0.9, 0.9],  # Out of allowed range
        "SliceThickness": 1.2  # Out of range
    }

def test_load_ref_pydantic_models(t1_mpr_dicom_values):
    # Path to the module file (e.g., "dcm_check/tests/ref_pydantic.py")
    module_path = "dcm_check/tests/ref_pydantic.py"

    # Load all the models from the module
    load_ref_pydantic(module_path, "T1_MPR")
    load_ref_pydantic(module_path, "T2w_SPC")
    load_ref_pydantic(module_path, "Diff_1k")

def test_load_ref_pydantic_invalid_module():
    """Test loading a Pydantic model from an invalid module."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    with pytest.raises(ValueError, match="Acquisition 'FAKE' is not defined in ACQUISITION_MODELS."):
        load_ref_pydantic(module_path, "FAKE")

def test_load_ref_pydantic_no_acquisition_models():
    """Test loading a Pydantic model from a module without ACQUISITION_MODELS."""
    module_path = "dcm_check/tests/ref_pydantic_no_models.py"
    with pytest.raises(ValueError, match="No ACQUISITION_MODELS found in the module 'dcm_check/tests/ref_pydantic_no_models.py'."):
        load_ref_pydantic(module_path, "T1_MPR")

def test_t1_mpr_compliance(t1_mpr_dicom_values):
    """Test compliance for valid T1_MPR values."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    t1_mpr_model = load_ref_pydantic(module_path, "T1_MPR")

    # Verify that the model was loaded correctly
    assert t1_mpr_model is not None
    assert t1_mpr_model.__name__ == "T1_MPR_Config"

    # Validate compliance
    assert is_compliant(t1_mpr_model, t1_mpr_dicom_values)

    # Test get_compliance_summary
    compliance_summary = get_compliance_summary(t1_mpr_model, t1_mpr_dicom_values)
    assert len(compliance_summary) == 0  # Should be no errors if values are valid

def test_t1_mpr_compliance_summary_invalid(invalid_t1_mpr_dicom_values):
    """Test compliance summary for invalid T1_MPR values."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    t1_mpr_model = load_ref_pydantic(module_path, "T1_MPR")
    
    # Validate non-compliance
    assert not is_compliant(t1_mpr_model, invalid_t1_mpr_dicom_values)

    # Get compliance summary for errors
    compliance_summary = get_compliance_summary(t1_mpr_model, invalid_t1_mpr_dicom_values)
    assert len(compliance_summary) > 0  # Expect some errors

    # Example check for specific errors
    error_params = [error["Parameter"] for error in compliance_summary]
    assert "SeriesDescription" in error_params  # Expect failure on SeriesDescription
    assert "MagneticFieldStrength" in error_params  # Out of range error
    assert "RepetitionTime" in error_params  # Repetition vs Echo rule
    assert "PixelSpacing" in error_params  # Out of range error
    assert "SliceThickness" in error_params  # Out of range error

def test_t1_mpr_repetition_vs_echo_rule(t1_mpr_dicom_values):
    """Test that the RepetitionTime and EchoTime rule is enforced."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    t1_mpr_model = load_ref_pydantic(module_path, "T1_MPR")
    
    # Modify RepetitionTime to break the 2x EchoTime rule
    t1_mpr_dicom_values["EchoTime"] = 3000  # Too low
    
    # Validate non-compliance
    assert not is_compliant(t1_mpr_model, t1_mpr_dicom_values)

    # Check compliance summary
    compliance_summary = get_compliance_summary(t1_mpr_model, t1_mpr_dicom_values)

    assert len(compliance_summary) > 0
    assert compliance_summary[0]["Parameter"] == "Model-Level Error"
    assert compliance_summary[0]["Expected"] == "RepetitionTime must be at least 2x EchoTime"
    assert compliance_summary[0]["Value"] == "N/A"

def test_diffusion_config_compliance():
    """Test DiffusionConfig compliance for a sample diffusion acquisition."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    diffusion_model = load_ref_pydantic(module_path, "Diff_1k")

    valid_diffusion_values = {
        "SeriesDescription": "Diff_1k",
        "bValue": 1000,
        "NumberOfDirections": 30,
        "PixelSpacing": [1.25, 1.25],
        "SliceThickness": 1.25
    }
    
    # Validate compliance
    assert is_compliant(diffusion_model, valid_diffusion_values)

def test_diffusion_config_non_compliance():
    """Test non-compliance for DiffusionConfig."""
    module_path = "dcm_check/tests/ref_pydantic.py"
    diffusion_model = load_ref_pydantic(module_path, "Diff_1k")

    invalid_diffusion_values = {
        "SeriesDescription": "wrong",  # Wrong prefix
        "bValue": 1200, # Out of range
        "NumberOfDirections": 4,  # Too few directions
        "PixelSpacing": [1.4, 1.4],  # Out of allowed range
        "SliceThickness": 1.4  # Out of range
    }
    
    # Validate non-compliance
    assert not is_compliant(diffusion_model, invalid_diffusion_values)

    # Check compliance summary
    compliance_summary = get_compliance_summary(diffusion_model, invalid_diffusion_values)
    print(compliance_summary)
    assert len(compliance_summary) == 6
    error_params = [error["Parameter"] for error in compliance_summary]
    assert "SeriesDescription" in error_params
    assert "NumberOfDirections" in error_params
    assert "PixelSpacing" in error_params
    assert "SliceThickness" in error_params

if __name__ == "__main__":
    pytest.main(["-v", __file__])




__version__ = "0.1.3"

from .compliance_check import \
    get_dicom_values, \
    load_dicom, \
    create_reference_model, \
    load_ref_json, \
    load_ref_dicom, \
    load_ref_pydantic, \
    get_compliance_summary, \
    is_compliant

from .dcm_gen_session import \
    generate_json_ref 
    
from .dcm_read_session import \
    calculate_field_score, \
    calculate_match_score, \
    find_closest_matches, \
    read_session, \
    interactive_mapping
    
from .dcm_check_session import \
    get_compliance_summaries_json



#!/usr/bin/env python

import numpy as np
import pytest

from pydantic import ValidationError
from pydicom.dataset import Dataset

from dcm_check import load_dicom, is_compliant, get_compliance_summary, load_ref_dicom, get_dicom_values
from dcm_check.tests.utils import create_empty_dicom

@pytest.fixture
def t1() -> Dataset:
    """Create a DICOM object with T1-weighted MRI metadata for testing."""

    ref_dicom = create_empty_dicom()
    
    # Set example attributes for T1-weighted MRI
    ref_dicom.SeriesDescription = "T1-weighted"
    ref_dicom.ProtocolName = "T1"
    ref_dicom.ScanningSequence = "GR"
    ref_dicom.SequenceVariant = "SP"
    ref_dicom.ScanOptions = "FS"
    ref_dicom.MRAcquisitionType = "3D"
    ref_dicom.RepetitionTime = "8.0"
    ref_dicom.EchoTime = "3.0"
    ref_dicom.InversionTime = "400.0"
    ref_dicom.FlipAngle = "15"
    ref_dicom.SAR = "0.1"
    ref_dicom.SliceThickness = "1.0"
    ref_dicom.SpacingBetweenSlices = "1.0"
    ref_dicom.PixelSpacing = ["0.5", "0.5"]
    ref_dicom.Rows = 256
    ref_dicom.Columns = 256
    ref_dicom.ImageOrientationPatient = ["1", "0", "0", "0", "1", "0"]
    ref_dicom.ImagePositionPatient = ["-128", "-128", "0"]
    ref_dicom.Laterality = "R"
    ref_dicom.PatientPosition = "HFS"
    ref_dicom.BodyPartExamined = "BRAIN"
    ref_dicom.PatientOrientation = ["A", "P", "R", "L"]
    ref_dicom.AcquisitionMatrix = [256, 0, 0, 256]
    ref_dicom.InPlanePhaseEncodingDirection = "ROW"
    ref_dicom.EchoTrainLength = 1
    ref_dicom.PercentPhaseFieldOfView = "100"
    ref_dicom.AcquisitionContrast = "UNKNOWN"
    ref_dicom.PixelBandwidth = "200"
    ref_dicom.DeviceSerialNumber = "12345"
    ref_dicom.ImageType = ["ORIGINAL", "PRIMARY", "M", "ND"]

    # Set PixelData to a 10x10 array of random integers
    ref_dicom.Rows = 10
    ref_dicom.Columns = 10
    ref_dicom.BitsAllocated = 16
    ref_dicom.BitsStored = 16
    ref_dicom.HighBit = 15
    ref_dicom.PixelRepresentation = 0
    ref_dicom.PixelData = np.random.randint(0, 2**16, (10, 10)).astype(np.uint16).tobytes()

    return ref_dicom

def test_load_dicom(tmp_path, t1):
    dicom_path = tmp_path / "ref_dicom.dcm"
    t1.save_as(dicom_path, enforce_file_format=True)
    dicom_values = load_dicom(dicom_path)
    assert dicom_values["SeriesDescription"] == "T1-weighted"

def test_dicom_compliance_identity(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)
    assert is_compliant(reference_model=reference, dicom_values=t1_values)
    assert len(get_compliance_summary(reference_model=reference, dicom_values=t1_values)) == 0

def test_dicom_compliance_specific_fields_compliant(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=["RepetitionTime", "EchoTime", "InversionTime"])
    t1_values["SliceThickness"] = 2.0
    assert is_compliant(reference_model=reference, dicom_values=t1_values)
    assert len(get_compliance_summary(reference_model=reference, dicom_values=t1_values)) == 0

def test_dicom_compliance_specific_fields_non_compliant(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=["RepetitionTime", "EchoTime", "InversionTime"])
    t1_values["RepetitionTime"] = 8.1

    assert not is_compliant(reference_model=reference, dicom_values=t1_values)

    compliance_summary = get_compliance_summary(reference_model=reference, dicom_values=t1_values)
    assert len(compliance_summary) == 1
    assert compliance_summary[0]["Parameter"] == "RepetitionTime"
    assert compliance_summary[0]["Expected"] == "8.0"
    assert compliance_summary[0]["Value"] == 8.1

def test_dicom_compliance_small_change(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)

    t1_values["RepetitionTime"] = 8.1

    assert not is_compliant(reference_model=reference, dicom_values=t1_values)

    compliance_summary = get_compliance_summary(reference_model=reference, dicom_values=t1_values)
    assert len(compliance_summary) == 1
    assert compliance_summary[0]["Parameter"] == "RepetitionTime"
    assert compliance_summary[0]["Expected"] == "8.0"
    assert compliance_summary[0]["Value"] == 8.1

def test_dicom_compliance_num_errors(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)

    t1_values["RepetitionTime"] = 8.1
    t1_values["EchoTime"] = 3.1
    t1_values["InversionTime"] = 400.1

    assert len(get_compliance_summary(reference_model=reference, dicom_values=t1_values)) == 3

def test_dicom_compliance_error_message(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)

    t1_values["RepetitionTime"] = 8.1
    t1_values["EchoTime"] = 3.1
    t1_values["InversionTime"] = 400.1

    errors = get_compliance_summary(reference_model=reference, dicom_values=t1_values)

    assert errors[0]["Parameter"] == "RepetitionTime"
    assert errors[1]["Parameter"] == "EchoTime"
    assert errors[2]["Parameter"] == "InversionTime"
    assert errors[0]["Expected"] == "8.0"
    assert errors[1]["Expected"] == "3.0"
    assert errors[2]["Expected"] == "400.0"
    assert errors[0]["Value"] == 8.1
    assert errors[1]["Value"] == 3.1
    assert errors[2]["Value"] == 400.1

def test_dicom_compliance_error_message_missing_field(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)

    del t1_values["RepetitionTime"]

    errors = get_compliance_summary(reference_model=reference, dicom_values=t1_values)
    compliant = is_compliant(reference_model=reference, dicom_values=t1_values)

    assert not compliant
    assert len(errors) == 1
    assert errors[0]["Parameter"] == "RepetitionTime"
    assert errors[0]["Expected"] == "Field required"
    assert errors[0]["Value"] == "N/A"

def test_dicom_compliance_error_raise(t1):
    t1_values = get_dicom_values(t1)
    reference = load_ref_dicom(t1_values, fields=None)

    t1_values["RepetitionTime"] = 8.1

    with pytest.raises(ValidationError):
        get_compliance_summary(reference_model=reference, dicom_values=t1_values, raise_errors=True)

def test_get_dicom_values_sequence(t1):
    t1.SequenceOfUltrasoundRegions = [Dataset(), Dataset()]
    t1.SequenceOfUltrasoundRegions[0].RegionLocationMinX0 = 0
    t1.SequenceOfUltrasoundRegions[0].RegionLocationMinY0 = 0
    t1.SequenceOfUltrasoundRegions[0].PhysicalUnitsXDirection = 1
    t1.SequenceOfUltrasoundRegions[0].PhysicalUnitsYDirection = 1
    t1.SequenceOfUltrasoundRegions[1].RegionLocationMinX0 = 0
    t1.SequenceOfUltrasoundRegions[1].RegionLocationMinY0 = 0
    t1.SequenceOfUltrasoundRegions[1].PhysicalUnitsXDirection = 1
    t1.SequenceOfUltrasoundRegions[1].PhysicalUnitsYDirection = 1

    dicom_values = get_dicom_values(t1)
    assert dicom_values["SequenceOfUltrasoundRegions"][0]["RegionLocationMinX0"] == 0
    assert dicom_values["SequenceOfUltrasoundRegions"][1]["RegionLocationMinY0"] == 0
    assert dicom_values["SequenceOfUltrasoundRegions"][0]["PhysicalUnitsXDirection"] == 1
    assert dicom_values["SequenceOfUltrasoundRegions"][1]["PhysicalUnitsYDirection"] == 1
    

if __name__ == "__main__":
    pytest.main(["-v", __file__])
    



#!/usr/bin/env python

import pytest
import json

from dcm_check import load_ref_json, load_dicom, get_compliance_summary, is_compliant
from dcm_check.tests.utils import create_empty_dicom
from pydantic_core import PydanticUndefined
from typing import Literal

@pytest.fixture
def dicom_test_file(tmp_path):
    """Fixture to create a DICOM file used as test input."""
    dicom_path = tmp_path / "ref_dicom.dcm"
    ds = create_empty_dicom()

    ds.EchoTime = 3.0
    ds.RepetitionTime = 8.0
    ds.SeriesDescription = "T1-weighted"
    ds.ImageType = ["ORIGINAL", "PRIMARY", "M", "ND"]

    ds.save_as(dicom_path, enforce_file_format=True)
    return str(dicom_path)

@pytest.fixture
def json_ref_no_dcm(tmp_path_factory):
    """Fixture to create a JSON reference file for testing."""
    test_json = {
        "acquisitions": {
            "T1": {
                "fields": [
                    {"field": "EchoTime", "tolerance": 0.1, "value": 3.0},
                    {"field": "RepetitionTime", "value": 8.0},
                    {"field": "SeriesDescription", "value": "*T1*"}
                ],
                "series": [
                    {
                        "name": "Series 1",
                        "fields": [
                            {"field": "ImageType", "contains": "M"}
                        ]
                    }
                ]
            }
        }
    }
    
    json_path = tmp_path_factory.mktemp("data") / "json_ref_no_dcm.json"
    with open(json_path, 'w') as f:
        json.dump(test_json, f)
    
    return str(json_path)

@pytest.fixture
def json_ref_with_dcm(tmp_path_factory, dicom_test_file):
    """Fixture to create a JSON reference file for testing."""
    test_json = {
        "acquisitions": {
            "T1": {
                "ref": dicom_test_file,
                "fields": [
                    {"field": "EchoTime", "tolerance": 0.1},
                    {"field": "RepetitionTime"},
                    {"field": "SeriesDescription"}
                ],
                "series": [
                    {
                        "name": "Series 1",
                        "fields": [
                            {"field": "ImageType", "contains": "M"}
                        ]
                    }
                ]
            }
        }
    }
    
    json_path = tmp_path_factory.mktemp("data") / "json_ref_with_dcm.json"
    with open(json_path, 'w') as f:
        json.dump(test_json, f)
    
    return str(json_path)

def test_load_ref_json(json_ref_no_dcm):
    """Test that JSON configuration can be loaded and generates a reference model."""
    reference_model = load_ref_json(json_path=json_ref_no_dcm, acquisition="T1", series_name="Series 1")

    # Verify that the model was created correctly with exact and pattern matching fields
    assert reference_model is not None
    assert "EchoTime" in reference_model.model_fields
    assert "RepetitionTime" in reference_model.model_fields
    assert "SeriesDescription" in reference_model.model_fields
    assert "ImageType" in reference_model.model_fields

    # Check EchoTime with tolerance
    assert reference_model.model_fields["EchoTime"].default == 3.0
    assert reference_model.model_fields["EchoTime"].metadata[1].ge == 2.9
    assert reference_model.model_fields["EchoTime"].metadata[1].le == 3.1

    # Check that RepetitionTime is required, with an exact match of 8.0
    assert reference_model.model_fields["RepetitionTime"].default is PydanticUndefined
    assert reference_model.model_fields["RepetitionTime"].annotation == Literal[8.0]

    # Check that pattern is correctly set on SeriesDescription using metadata
    assert reference_model.model_fields["SeriesDescription"].metadata[0].pattern == ".*T1.*"

def test_json_compliance_within_tolerance_with_dcm(json_ref_with_dcm, dicom_test_file):
    """Test compliance when values are within tolerance for JSON configuration with series."""
    t1_dicom_values = load_dicom(dicom_test_file)
    reference_model = load_ref_json(json_path=json_ref_with_dcm, acquisition="T1", series_name="Series 1")

    # Adjust EchoTime within tolerance (original value is 3.0, tolerance 0.1)
    t1_dicom_values["EchoTime"] = 3.05
    compliance_summary = get_compliance_summary(reference_model, t1_dicom_values)

    assert is_compliant(reference_model, t1_dicom_values)
    assert len(compliance_summary) == 0

def test_json_compliance_outside_tolerance_with_dcm(json_ref_with_dcm, dicom_test_file):
    """Test compliance when values exceed tolerance for JSON configuration with series."""
    t1_dicom_values = load_dicom(dicom_test_file)
    reference_model = load_ref_json(json_path=json_ref_with_dcm, acquisition="T1", series_name="Series 1")

    # Adjust EchoTime beyond tolerance (original value is 3.0, tolerance 0.1)
    t1_dicom_values["EchoTime"] = 3.2
    compliance_summary = get_compliance_summary(reference_model, t1_dicom_values)
    assert len(compliance_summary) == 1
    assert compliance_summary[0]["Parameter"] == "EchoTime"
    assert compliance_summary[0]["Expected"] == "Input should be less than or equal to 3.1"
    assert compliance_summary[0]["Value"] == 3.2

def test_json_compliance_pattern_match(json_ref_no_dcm, dicom_test_file):
    """Test compliance with a pattern match for SeriesDescription within series."""
    t1_dicom_values = load_dicom(dicom_test_file)
    reference_model = load_ref_json(json_path=json_ref_no_dcm, acquisition="T1", series_name="Series 1")

    # Change SeriesDescription to match pattern "*T1*"
    t1_dicom_values["SeriesDescription"] = "Another_T1_Sequence"
    compliance_summary = get_compliance_summary(reference_model, t1_dicom_values)
    assert len(compliance_summary) == 0  # Should pass pattern match

if __name__ == "__main__":
    pytest.main(["-v", __file__])



import os
import json
import pandas as pd
import argparse
import re

from scipy.optimize import linear_sum_assignment
from dcm_check import load_dicom

try:
    import curses
except ImportError:
    curses = None

MAX_DIFF_SCORE = 10  # Maximum allowed difference score for each field to avoid unmanageably large values

def levenshtein_distance(s1, s2):
    """
    Calculate the Levenshtein distance between two strings.
    """
    if len(s1) < len(s2):
        return levenshtein_distance(s2, s1)

    # Initialize a row with incremental values [0, 1, 2, ..., len(s2)]
    previous_row = range(len(s2) + 1)

    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]

def calculate_field_score(expected, actual, tolerance=None, contains=None):
    """Calculate the difference between expected and actual values, with caps for large scores."""
    if isinstance(expected, str) and ("*" in expected or "?" in expected):
        pattern = re.compile("^" + expected.replace("*", ".*").replace("?", ".") + "$")
        if pattern.match(actual):
            return 0  # Pattern matched, no difference
        return min(MAX_DIFF_SCORE, 5)  # Pattern did not match, fixed penalty

    if contains:
        if (isinstance(actual, str) and contains in actual) or (isinstance(actual, (list, tuple)) and contains in actual):
            return 0  # Contains requirement fulfilled, no difference
        return min(MAX_DIFF_SCORE, 5)  # 'Contains' not met, fixed penalty

    if isinstance(expected, (list, tuple)) or isinstance(actual, (list, tuple)):
        expected_tuple = tuple(expected) if not isinstance(expected, tuple) else expected
        actual_tuple = tuple(actual) if not isinstance(actual, tuple) else actual
        
        if all(isinstance(e, (int, float)) for e in expected_tuple) and all(isinstance(a, (int, float)) for a in actual_tuple) and len(expected_tuple) == len(actual_tuple):
            if tolerance is not None:
                return min(MAX_DIFF_SCORE, sum(abs(e - a) for e, a in zip(expected_tuple, actual_tuple) if abs(e - a) > tolerance))

        max_length = max(len(expected_tuple), len(actual_tuple))
        expected_padded = expected_tuple + ("",) * (max_length - len(expected_tuple))
        actual_padded = actual_tuple + ("",) * (max_length - len(actual_tuple))
        return min(MAX_DIFF_SCORE, sum(levenshtein_distance(str(e), str(a)) for e, a in zip(expected_padded, actual_padded)))
    
    if isinstance(expected, (int, float)) and isinstance(actual, (int, float)):
        if tolerance is not None:
            if abs(expected - actual) <= tolerance:
                return 0
        return min(MAX_DIFF_SCORE, abs(expected - actual))
    
    return min(MAX_DIFF_SCORE, levenshtein_distance(str(expected), str(actual)))

def calculate_match_score(acquisition, series_row):
    """Calculate the capped total difference score between an acquisition and a DICOM entry."""
    diff_score = 0.0
    for field, expected_value in acquisition["fields"].items():
        actual_value = series_row.get(field, "N/A")
        tolerance = acquisition.get("tolerance", {}).get(field)
        contains = acquisition.get("contains", {}).get(field)
        diff = calculate_field_score(expected_value, actual_value, tolerance=tolerance, contains=contains)
        diff_score += diff
    return round(diff_score, 2)

def find_closest_matches(session_df, acquisitions_info):
    """Compute minimal score assignments for acquisitions, handling unassigned rows."""
    cost_matrix = []
    possible_assignments = []

    for i, row in session_df.iterrows():
        row_costs = []
        row_assignments = []
        
        for acq_info in acquisitions_info:
            acq_name = acq_info["name"]
            acq_match_score = calculate_match_score(acq_info, row)

            if not acq_info["series"]:  # Acquisitions without groups (assign group as None)
                row_costs.append(acq_match_score)
                row_assignments.append((i, acq_name, None, acq_match_score))
            else:
                for series in acq_info["series"]:
                    series_name = series["name"]
                    series_diff_score = calculate_match_score(series, row)
                    total_score = acq_match_score + series_diff_score
                    row_costs.append(total_score)
                    row_assignments.append((i, acq_name, series_name, total_score))

        cost_matrix.append(row_costs)
        possible_assignments.append(row_assignments)

    cost_matrix = pd.DataFrame(cost_matrix)
    row_indices, col_indices = linear_sum_assignment(cost_matrix)

    best_acquisitions = [None] * len(session_df)
    best_series = [None] * len(session_df)
    best_scores = [None] * len(session_df)  # Use NaN for unmatched scores

    for row_idx, col_idx in zip(row_indices, col_indices):
        _, acq_name, series_name, score = possible_assignments[row_idx][col_idx]
        best_acquisitions[row_idx] = acq_name
        best_series[row_idx] = series_name
        best_scores[row_idx] = score if acq_name else None  # Only assign score if acquisition is matched

    return best_acquisitions, best_series, best_scores

def read_session(reference_json, session_dir, return_acquisitions_info=False):
    """
    Read a DICOM session directory and map it to the closest acquisitions and series in a reference JSON file.

    Args:
        reference_json (str): Path to the JSON reference file.
        session_dir (str): Directory path containing DICOM files for the session.
        return_acquisitions_info (bool): If True, returns acquisitions_info for dynamic score recalculation.

    Returns:
        pd.DataFrame: DataFrame containing matched acquisitions and series with scores.
        acquisitions_info (optional): List of acquisitions and series info, used for score recalculation.
    """
    with open(reference_json, 'r') as f:
        reference_data = json.load(f)
    
    # Load acquisition and series configurations from JSON
    acquisitions_info = [
        {
            "name": acq_name,
            "fields": {field["field"]: field.get("value", field.get("contains")) for field in acquisition.get("fields", [])},
            "tolerance": {field["field"]: field["tolerance"] for field in acquisition.get("fields", []) if "tolerance" in field},
            "contains": {field["field"]: field["contains"] for field in acquisition.get("fields", []) if "contains" in field},
            "series": [
                {
                    "name": series["name"], 
                    "fields": {field["field"]: field.get("value", field.get("contains")) for field in series.get("fields", [])},
                    "tolerance": {field["field"]: field["tolerance"] for field in series.get("fields", []) if "tolerance" in field},
                    "contains": {field["field"]: field["contains"] for field in series.get("fields", []) if "contains" in field}
                } 
                for series in acquisition.get("series", [])
            ]
        }
        for acq_name, acquisition in reference_data.get("acquisitions", {}).items()
    ]
    
    # Collect all unique fields across acquisitions and series
    all_fields = {field for acq in acquisitions_info for field in acq["fields"].keys()}
    all_fields.update({field for acq in acquisitions_info for series in acq["series"] for field in series["fields"].keys()})

    session_data = []
    for root, _, files in os.walk(session_dir):
        for file in files:
            if file.endswith((".dcm", ".IMA")):
                dicom_path = os.path.join(root, file)
                dicom_values = load_dicom(dicom_path)
                
                dicom_entry = {field: tuple(dicom_values[field]) if isinstance(dicom_values.get(field), list) 
                               else dicom_values.get(field, "N/A") 
                               for field in all_fields}
                
                dicom_entry["DICOM_Path"] = dicom_path  # Store the DICOM path
                dicom_entry["InstanceNumber"] = int(dicom_values.get("InstanceNumber", 0))  # Add InstanceNumber if available

                session_data.append(dicom_entry)

    session_df = pd.DataFrame(session_data)

    # Sort by InstanceNumber or DICOM_Path
    if "InstanceNumber" in session_df.columns:
        session_df.sort_values("InstanceNumber", inplace=True)
    else:
        session_df.sort_values("DICOM_Path", inplace=True)
    
    # Group by unique series fields and calculate 'First_DICOM' and 'Count'
    dedup_fields = all_fields
    series_count_df = (
        session_df.groupby(list(dedup_fields))
        .agg(First_DICOM=('DICOM_Path', 'first'), Count=('DICOM_Path', 'size'))
        .reset_index()
    )

    # Find closest matches for each series and calculate scores
    acquisitions, series, scores = find_closest_matches(series_count_df, acquisitions_info)

    series_count_df["Acquisition"] = acquisitions
    series_count_df["Series"] = series
    series_count_df["Match_Score"] = scores


    # Order headers with fields associated with acquisition_info first, then fields associated with series
    acquisition_fields = {field for acq in acquisitions_info for field in acq["fields"].keys()}
    series_fields = {field for acq in acquisitions_info for series in acq["series"] for field in series["fields"].keys() if field not in acquisition_fields}
    ordered_headers = list(acquisition_fields) + list(series_fields) + ["First_DICOM", "Count", "Acquisition", "Series", "Match_Score"]
    series_count_df = series_count_df[ordered_headers]
    series_count_df.sort_values(["Acquisition", "Series", "Match_Score"], inplace=True)

    # Return the DataFrame and optionally acquisitions_info
    if return_acquisitions_info:
        return series_count_df, acquisitions_info
    return series_count_df

def interactive_mapping(df, acquisitions_info):
    """
    Launch an interactive CLI for adjusting acquisition mappings with dynamic match score updates.
    """
    if not curses:
        raise ImportError("curses module is not available. Please install it to use interactive mode.")
    
    def calculate_column_widths(df, padding=2):
        column_widths = {}
        for col in df.columns:
            max_content_width = max(len(str(x)) for x in df[col]) if len(df) > 0 else 10
            column_widths[col] = max(len(col), max_content_width) + padding
        return column_widths

    def draw_menu(stdscr, df, highlighted_row, column_widths, selected_values):
        stdscr.clear()
        h, w = stdscr.getmaxyx()  # Get the screen height and width

        # Calculate total row height and truncate rows if needed
        max_visible_rows = h - 2  # Leave space for header row and bottom navigation

        # Calculate column widths and truncate if they exceed screen width
        available_width = w - 2  # Start with screen width, adjusted for padding
        truncated_column_widths = {}
        for col_name in df.columns:
            # skip First_DICOM and Count columns
            if col_name in ["First_DICOM", "Count"]:
                continue
            col_width = min(column_widths[col_name], available_width)
            truncated_column_widths[col_name] = col_width
            available_width -= col_width
            if available_width <= 0:
                break  # No more space left for additional columns

        # Draw headers
        x = 2
        for col_name in truncated_column_widths.keys():
            header_text = col_name.ljust(truncated_column_widths[col_name])[:truncated_column_widths[col_name]]
            stdscr.addstr(1, x, header_text)
            x += truncated_column_widths[col_name]

        # Draw rows with Acquisition and Series columns highlighted
        visible_rows = df.iloc[:max_visible_rows]  # Limit to max visible rows
        for idx, row in visible_rows.iterrows():
            y = idx + 2
            x = 2
            for col_name in truncated_column_widths.keys():
                is_selected_column = col_name in ["Acquisition", "Series"] and idx == highlighted_row
                cell_text = str(selected_values[col_name] if is_selected_column and selected_values is not None else row[col_name]).ljust(truncated_column_widths[col_name])[:truncated_column_widths[col_name]]

                if is_selected_column:
                    stdscr.attron(curses.A_REVERSE)
                    stdscr.addstr(y, x, cell_text)
                    stdscr.attroff(curses.A_REVERSE)
                else:
                    stdscr.addstr(y, x, cell_text)
                x += truncated_column_widths[col_name]
        stdscr.refresh()


    def recalculate_match_score(row_idx, df, acquisitions_info):
        acquisition_name = df.at[row_idx, "Acquisition"]
        series_name = df.at[row_idx, "Series"]

        if pd.notna(acquisition_name):
            acquisition_info = next((acq for acq in acquisitions_info if acq["name"] == acquisition_name), None)
            if acquisition_info:
                if pd.notna(series_name):
                    series_info = next((series for series in acquisition_info["series"] if series["name"] == series_name), None)
                    if series_info:
                        score = calculate_match_score(acquisition_info, df.loc[row_idx]) + calculate_match_score(series_info, df.loc[row_idx])
                    else:
                        score = float('inf')
                else:
                    score = calculate_match_score(acquisition_info, df.loc[row_idx])
                return score
        return float('inf')
    
    def interactive_loop(stdscr, df):
        curses.curs_set(0)  # Hide the cursor
        highlighted_row = 0
        last_highlighted_row = 0
        selected_row = None
        selected_values = None
        column_widths = calculate_column_widths(df)

        while True:
            draw_menu(stdscr, df, highlighted_row, column_widths, selected_values)
            key = stdscr.getch()

            if key in [curses.KEY_UP, curses.KEY_DOWN]:
                # Store the last highlighted row before moving
                last_highlighted_row = highlighted_row
                highlighted_row += -1 if key == curses.KEY_UP else 1
                highlighted_row = max(0, min(len(df) - 1, highlighted_row))

                # If we're moving a selected assignment, perform a dynamic swap and recalculate scores
                if selected_values is not None:
                    # Swap values between the last and current highlighted rows
                    df.loc[last_highlighted_row, ["Acquisition", "Series"]], df.loc[highlighted_row, ["Acquisition", "Series"]] = (
                        df.loc[highlighted_row, ["Acquisition", "Series"]].values,
                        selected_values.values()
                    )
                    
                    # Recalculate and update match scores for both swapped rows
                    df.at[last_highlighted_row, "Match_Score"] = recalculate_match_score(last_highlighted_row, df, acquisitions_info)
                    df.at[highlighted_row, "Match_Score"] = recalculate_match_score(highlighted_row, df, acquisitions_info)

            elif key == 10:  # Enter key
                if selected_row is None:
                    # Start moving the selected assignment
                    selected_row = highlighted_row
                    selected_values = df.loc[selected_row, ["Acquisition", "Series"]].to_dict()
                    # Clear original position
                    df.loc[selected_row, ["Acquisition", "Series"]] = None
                else:
                    # Place assignment at the new position and deselect
                    df.loc[highlighted_row, ["Acquisition", "Series"]] = pd.Series(selected_values)
                    df.at[highlighted_row, "Match_Score"] = recalculate_match_score(highlighted_row, df, acquisitions_info)
                    selected_row = None
                    selected_values = None  # Reset selected values

            elif key == 27:  # ESC key
                # Exit the interactive loop
                break

    curses.wrapper(interactive_loop, df)
    return df

def main():
    parser = argparse.ArgumentParser(description="Map a DICOM session directory to a JSON reference file and print the closest acquisition and series matches.")
    parser.add_argument("--ref", required=True, help="Path to the JSON reference file.")
    parser.add_argument("--session_dir", required=True, help="Directory containing DICOM files for the session.")
    args = parser.parse_args()
    
    df, acquisitions_info = read_session(args.ref, args.session_dir, return_acquisitions_info=True)  # Adjusted read_session to return acquisitions_info
    # Drop First_DICOM and Count for display
    df.drop(columns=["First_DICOM", "Count"], inplace=True)

    # Interactive session
    print("Entering interactive mapping mode. Use arrow keys to navigate, Enter to select and move, and Esc to finish.")
    adjusted_df = interactive_mapping(df, acquisitions_info)
    
    # Display the adjusted dataframe
    print(adjusted_df)

if __name__ == "__main__":
    main()